"""Plugin for Cambium specific configuration.

This plugin is only applicable for ReEDs, but could work with similarly arrange data
"""

from argparse import ArgumentParser

import polars as pl
from loguru import logger

from r2x.api import System
from r2x.config_scenario import Scenario
from r2x.models.branch import MonitoredLine
from r2x.models.core import MinMax
from r2x.models.generators import Generator
from r2x.models.topology import ACBus
from r2x.parser.handler import BaseParser


def cli_arguments(parser: ArgumentParser):
    """CLI arguments for the plugin."""
    parser.add_argument(
        "--perturb",
        type=float,
        help="Load perturbation scalar",
    )


def update_system(
    config: Scenario,
    system: System,
    perturb: float,
    parser: BaseParser | None = None,
) -> System:
    """Apply hurdle rate between regions.

    This function updates the default hurdle rate for the ReEDS parser using a new file

    Parameters
    ----------
    config : Scenario
        The scenario configuration.
    system : System
        The system object to be updated.
    hurdle_rate : str, optional
        The hurdle rate to apply betwen regions.
    parser : BaseParser, optional
        The parser object used for parsing.
    """
    if not config.output_model == "plexos" or not config.input_model == "reeds-US":
        msg = "Plugin `cambium.py` is not compatible with a model that is not Plexos."
        raise NotImplementedError(msg)

    logger.info("Applying cambium configuration")

    # Removing probabilistic outages and apply derate instead.
    system = _derate_plants(system)

    # Add Fixed load for certain firm technologies.
    for generator in system.get_components(
        Generator,
        filter_func=lambda x: any(tech in x.ext["reeds_tech"] for tech in ["nuclear", "lfill", "biopower"]),
    ):
        generator.ext["Fixed Load"] = generator.active_power

    # Add load perturb scalar.
    for zone in system.get_components(ACBus):
        zone.ext["Load Scalar"] = perturb

    # comment out rest of this for not including hurdle rates
    if parser is not None:
        if not (parser.data.get("hurdle_rate") is not None):
            msg = "Hurdle rate file not found. Skipping pluging."
            logger.debug(msg)
            return system
        hurdle_rate_data = parser.data["hurdle_rate"]
        for line in system.get_components(MonitoredLine):
            to_bus = line.to_bus.name
            from_bus = line.from_bus.name
            hurdle_rate = (
                hurdle_rate_data.filter(pl.col("from_bus") == from_bus)
                .filter(pl.col("to_bus") == to_bus)["hurdle_rate"]
                .item()
            )
            logger.debug(f"Setting hurdle rate between {to_bus=} and {from_bus=} of {hurdle_rate=}")
            if to_bus != from_bus:
                if previous_hurdle := line.ext.get("Wheeling Charge"):
                    logger.debug(
                        "Changing hurdle rate for {} from {} to {}.",
                        line.name,
                        previous_hurdle,
                        hurdle_rate,
                    )
                # NOTE: This assume that we have the same hurdle rate in both directions.
                line.ext["Wheeling Charge"] = hurdle_rate
                line.ext["Wheeling Charge Back"] = hurdle_rate
    return system


def _derate_plants(system):
    for generator in system.get_components(Generator):
        if "distpv" in generator.name:
            generator.planned_outage_rate = None
        if generator.planned_outage_rate is None or generator.forced_outage_rate is None:
            continue

        generator.active_power = (
            (1 - generator.planned_outage_rate) * (1 - generator.forced_outage_rate) * generator.active_power
        )
        if hasattr(generator, "active_power_limits") and generator.active_power_limits:
            min_max_ratio = (
                generator.active_power_limits.max - generator.active_power_limits.min
            ) / generator.active_power_limits.max
            generator.active_power_limits = MinMax(
                min=generator.active_power_limits.min * min_max_ratio, max=generator.active_power
            )
        generator.planned_outage_rate = None
        generator.forced_outage_rate = None
        generator.mean_time_to_repair = None
    return system



"""Compilation of polars function that we use in the parsing process."""

from datetime import datetime
from typing import Any

import polars as pl
from loguru import logger
from polars.lazyframe import LazyFrame

from r2x.parser.plexos_utils import DATAFILE_COLUMNS


def pl_filter_year(
    data: pl.DataFrame, year: int | None = None, year_columns: list[str] = ["t", "year"], **kwargs
) -> pl.DataFrame:
    """Filter the DataFrame by a specific year.

    Parameters
    ----------
    df : pl.DataFrame
        The DataFrame to filter.
    year : int | None, optional
        The year to filter by, default is None.
    year_columns : list[str], optional
        The columns to check for year filtering, by default ['t', 'year'].
    **kwargs : dict, optional
        Additional arguments, can contain 'solve_year' to override the year.

    Returns
    -------
    pl.DataFrame
        The filtered DataFrame.

    Raises
    ------
    KeyError
        If more than one column is identified as year.
    """
    if year is None and kwargs.get("solve_year"):
        year = kwargs["solve_year"]

    if year is None:
        return data

    matching_names = list(set(year_columns).intersection(data.collect_schema()))
    if not matching_names:
        return data

    if len(matching_names) > 1:
        raise KeyError(f"More than one column identified as year. {matching_names=}")
    logger.trace("Filtering data for year {}", year)
    return data.filter(pl.col(matching_names[0]) == year)


def pl_remove_duplicates(data: pl.DataFrame, columns: DATAFILE_COLUMNS | list[str]) -> pl.DataFrame:
    """Remove duplicate rows from the DataFrame based on certain columns.

    Parameters
    ----------
    data : pl.DataFrame
        The DataFrame from which to remove duplicates.
    columns : DATAFILE_COLUMNS | list[str]
        The column type enumeration or list of strings used to determine columns for duplication check.

    Returns
    -------
    pl.DataFrame
        DataFrame without duplicate rows.
    """
    columns = columns.value if isinstance(columns, DATAFILE_COLUMNS) else columns
    columns_to_check = [
        col
        for col in columns
        if col in ["name", "pattern", "year", "DateTime", "month", "day", "period", "hour"]
    ]

    if not data.filter(data.select(columns_to_check).is_duplicated()).is_empty():
        logger.warning("File has duplicated rows. Removing duplicates.")
        data = data.unique(subset=columns_to_check).sort(pl.all())

    return data


def pl_lowercase(data: pl.DataFrame, **kwargs):
    """Convert all string columns to lowercase.

    Parameters
    ----------
    df : pl.DataFrame
        The DataFrame with columns to be lowercased.
    **kwargs : dict, optional
        Additional arguments.

    Returns
    -------
    pl.DataFrame
        The DataFrame with lowercase string columns.
    """
    logger.trace("Lowercase columns: {}", data.collect_schema().names())
    result = data.with_columns(pl.col(pl.String).str.to_lowercase()).rename(
        {column: column.lower() for column in data.collect_schema().names()}
    )
    logger.trace("New columns: {}", data.collect_schema().names())
    return result


def pl_rename(
    data: pl.DataFrame,
    column_mapping: dict[str, str] | None = None,
    **kwargs,
):
    """Rename columns in the DataFrame using a provided column mapping.

    Parameters
    ----------
    df : pl.DataFrame
        The DataFrame to rename columns.
    column_mapping : dict[str, str] | None, optional
        Mapping of original column names to new column names, by default None.
    **kwargs : dict, optional
        Additional arguments.

    Returns
    -------
    pl.DataFrame
        The DataFrame with renamed columns.
    """
    if not column_mapping:
        logger.trace("No column mapping provided")
        return data
    logger.trace("Apply column mapping {}", column_mapping)
    return data.select(pl.col(column_mapping.keys())).select(
        pl.all().name.map(lambda col_name: column_mapping.get(col_name, col_name))
    )


def pl_left_multi_join(l_df: pl.LazyFrame, *r_dfs: pl.DataFrame, **kwargs):
    """Perform a left join on multiple DataFrames.

    Parameters
    ----------
    l_df : pl.LazyFrame
        The left DataFrame for joining.
    *r_dfs : pl.LazyFrame
        The right DataFrames to join.
    **kwargs : dict, optional
        Additional arguments.

    Returns
    -------
    pl.DataFrame
        The resulting DataFrame after the join.

    Raises
    ------
    AssertionError
        If the merge results in fewer rows than the left DataFrame.
    """
    original_keys = set(l_df.collect_schema().names())
    output_df = l_df
    if not r_dfs:
        return l_df

    current_keys: set[Any] = set()
    for r_df in r_dfs:
        current_keys = set(r_df.collect_schema().names())
        current_keys = original_keys.intersection(current_keys)
        if isinstance(r_df, LazyFrame):
            r_df = r_df.collect()
        output_df = output_df.join(r_df, on=list(current_keys), how="left", coalesce=True)

    if isinstance(output_df, pl.LazyFrame):
        output_df = output_df.collect()

    l_df_shape = l_df.collect().shape[0] if isinstance(l_df, pl.LazyFrame) else l_df.shape[0]
    assert output_df.shape[0] == l_df_shape, (
        f"Merge resulted in less rows. Check the shared keys. {original_keys=} vs {current_keys=}"
    )
    return output_df


def pl_create_date_range(year: int, interval: str = "1h"):
    """Create a DataFrame with a date range for the given year.

    Parameters
    ----------
    year : int
        The year to create the date range for.
    interval : str, optional
        Time interval for the date range, by default '1h'.

    Returns
    -------
    pl.DataFrame
        DataFrame with datetime, year, month, day, and hour columns.
    """
    dt = pl.datetime_range(
        datetime(year, 1, 1), datetime(year + 1, 1, 1), interval, eager=True, closed="left"
    ).alias("datetime")
    date_df = pl.DataFrame({"datetime": dt})
    date_df = date_df.with_columns(
        [
            date_df["datetime"].dt.year().alias("year").cast(pl.Int64),
            date_df["datetime"].dt.month().alias("month").cast(pl.Int64),
            date_df["datetime"].dt.day().alias("day").cast(pl.Int64),
            date_df["datetime"].dt.hour().alias("hour").cast(pl.Int64),
        ]
    )
    return date_df



"""Script that creates simple 2-area pjm systems for testing."""

import pathlib
from datetime import datetime, timedelta

from infrasys.cost_curves import CostCurve, UnitSystem
from infrasys.time_series_models import SingleTimeSeries
from infrasys.value_curves import LinearCurve

from r2x.api import System
from r2x.config_scenario import get_scenario_configuration
from r2x.enums import ACBusTypes, PrimeMoversType, ReserveDirection, ReserveType, ThermalFuels
from r2x.models.branch import AreaInterchange, Line, MonitoredLine
from r2x.models.core import FromTo_ToFrom, MinMax, ReserveMap, UpDown
from r2x.models.costs import ThermalGenerationCost
from r2x.models.generators import RenewableDispatch, ThermalStandard
from r2x.models.load import PowerLoad
from r2x.models.services import Reserve
from r2x.models.topology import ACBus, Area, LoadZone
from r2x.runner import run_exporter
from r2x.units import ActivePower, Percentage, Time, Voltage, ureg
from r2x.utils import get_enum_from_string, read_json


def pjm_2area() -> System:
    """Return the PJM 2-area test system."""
    fpath = pathlib.Path(__file__).parent.parent  # So it points to tests
    fname = "data/pjm_2area_data.json"
    pjm_2area_components = read_json(str(fpath / fname))

    system = System(
        name="pjm 2-area system", auto_add_composed_components=True, description="Test system for PJM"
    )

    # Add topology elements
    system.add_component(Area(name="init"))
    system.add_component(Area(name="Area2"))
    system.add_component(LoadZone(name="init"))
    system.add_component(LoadZone(name="Area2"))

    for bus in pjm_2area_components["bus"]:
        system.add_component(
            ACBus(
                name=bus["name"],
                number=bus["number"],
                bustype=ACBusTypes(bus["bustype"]),
                base_voltage=Voltage(bus["base_voltage"], "kV"),
                area=system.get_component(Area, bus["area"]),
                load_zone=system.get_component(LoadZone, bus["area"]),
                magnitude=bus["magnitude"],
            )
        )

    # Add branches
    for branch in pjm_2area_components["branch"]:
        busf = system.get_component(ACBus, branch["FromBus"])
        bust = system.get_component(ACBus, branch["ToBus"])
        system.add_component(
            Line(
                name=branch["Name"],
                from_bus=busf,
                to_bus=bust,
                x=branch["x"],
                r=branch["r"],
                rating=branch["MaxRating"] * ureg.MW,
                active_power_flow=0.0,
                reactive_power_flow=0.0,
                angle_limits=MinMax(min=-0.7, max=0.7),
            )
        )
    # Add MonitoredLine
    busf = system.get_component(ACBus, "Bus_nodeC_1")
    bust = system.get_component(ACBus, "Bus_nodeC_2")
    branch_monitored = MonitoredLine(
        name="inter_area_line",
        from_bus=busf,
        r=0.003,
        x=0.03,
        to_bus=bust,
        rating_up=1000 * ureg.MW,
        rating_down=-1000 * ureg.MW,
    )
    system.add_component(branch_monitored)

    # Add area interchange
    system.add_component(
        AreaInterchange(
            name="1_2",
            flow_limits=FromTo_ToFrom(from_to=-150, to_from=150),
            from_area=system.get_component(Area, "init"),
            to_area=system.get_component(Area, "Area2"),
            active_power_flow=0.0,
        )
    )

    # Add thermal generators
    for gen in pjm_2area_components["thermal"]:
        system.add_component(
            ThermalStandard(
                name=gen["Name"],
                fuel=get_enum_from_string(gen["fuel"].lower(), ThermalFuels),
                prime_mover_type=PrimeMoversType.ST,
                unit_type=PrimeMoversType.ST,
                active_power=ActivePower(100, "MW"),
                min_down_time=Time(gen["MinTimeDn"], "hour"),
                min_up_time=Time(gen["MinTimeUp"], "hour"),
                mean_time_to_repair=Time(10.0, "hour"),
                forced_outage_rate=Percentage(0.0),
                planned_outage_rate=Percentage(0.0),
                ramp_limits=UpDown(up=gen["RampLimitsUp"], down=gen["RampLimitsDn"]),
                bus=system.get_component(ACBus, gen["BusName"]),
                category="thermal",
                operation_cost=ThermalGenerationCost(
                    shut_down=gen["ShutDnCost"],
                    start_up=gen["StartupCost"],
                    variable=CostCurve(
                        value_curve=LinearCurve(14.0),
                        power_units=UnitSystem.NATURAL_UNITS,
                        vom_cost=LinearCurve(gen["VOM"]),
                    ),
                ),
            )
        )

    # Solar generators
    solar_pv_01 = RenewableDispatch(
        name="PVBus5",
        bus=system.get_component(ACBus, "Bus_nodeC_1"),
        prime_mover_type=PrimeMoversType.PVe,
        unit_type=PrimeMoversType.PVe,
        active_power=384 * ureg.MW,
        category="solar",
    )
    system.add_component(solar_pv_01)

    wind_01 = RenewableDispatch(
        name="WindBus1",
        bus=system.get_component(ACBus, "Bus_nodeA_2"),
        prime_mover_type=PrimeMoversType.WT,
        unit_type=PrimeMoversType.WT,
        active_power=451.0 * ureg.MW,
        category="wind",
    )
    system.add_component(wind_01)

    # Add renewable profiles
    initial_time = datetime(year=2024, month=1, day=1)
    solar_array = pjm_2area_components["solar_ts"]  # Data is hourly resolution
    wind_array = pjm_2area_components["wind_ts"]  # Data is hourly resolution
    solar_ts = SingleTimeSeries.from_array(
        data=solar_array,
        initial_time=initial_time,
        resolution=timedelta(hours=1),
        variable_name="max_active_power",
    )
    wind_ts = SingleTimeSeries.from_array(
        data=wind_array,
        initial_time=initial_time,
        resolution=timedelta(hours=1),
        variable_name="rated_capacity",
    )
    #
    system.add_time_series(solar_ts, solar_pv_01)
    system.add_time_series(wind_ts, wind_01)

    initial_time = datetime(year=2024, month=1, day=1)
    for load in pjm_2area_components["load"]:
        load_component = PowerLoad(
            name=load["BusName"],
            bus=system.get_component(ACBus, load["BusName"]),
            max_active_power=load["MaxLoad"] * ureg.MW,
        )
        ld_ts = SingleTimeSeries.from_array(
            data=load["ts"],
            initial_time=initial_time,
            resolution=timedelta(hours=1),
            variable_name="max_active_power",
        )
        system.add_component(load_component)
        system.add_time_series(ld_ts, load_component)

    # Create reserve
    reserve_map = ReserveMap(name="pjm_reserve_map")
    reserve = Reserve(
        name="SpinUp-pjm",
        region=system.get_component(LoadZone, "init"),
        reserve_type=ReserveType.SPINNING,
        vors=0.05,
        duration=3600.0,
        load_risk=0.5,
        time_frame=3600,
        direction=ReserveDirection.UP,
    )
    reserve_map.mapping[ReserveType.SPINNING.name].append(wind_01.name)
    reserve_map.mapping[ReserveType.SPINNING.name].append(solar_pv_01.name)
    system.add_components(reserve, reserve_map)

    return system


def pjm_2area_to_plexos(output_folder: pathlib.Path):
    system = pjm_2area()

    scenario = get_scenario_configuration(
        cli_args={"name": "pjm_2area"},
        user_dict={
            "input_model": "infrasys",
            "output_model": "plexos",
            "output_folder": output_folder,
            "model_year": 2024,
        },
    )
    run_exporter(scenario["pjm_2area"], system)



import json
from pathlib import Path
from tempfile import NamedTemporaryFile

import pytest
import yaml

from r2x.utils import check_file_exists, haskey, override_dict, read_user_dict


@pytest.mark.utils
@pytest.mark.parametrize(
    "original_dict, path, expected",
    [
        ({"a": {"b": {"c": 1}}}, ["a", "b", "c"], True),  # valid path
        ({"a": {"b": {"c": 1}}}, ["a", "b"], True),  # valid partial path
        ({"a": {"b": {"c": 1}}}, ["a", "x"], False),  # invalid key in path
        ({"a": {"b": {"c": 1}}}, ["a", "b", "d"], False),  # invalid path at last level
        ({}, ["a"], False),  # empty dictionary
        ({"a": {"b": {"c": 1}}}, ["a", "b", "c", "d"], False),  # path too long
    ],
)
def test_haskey(original_dict, path, expected):
    assert haskey(original_dict, path) == expected


def test_read_user_dict(tmp_path):
    user_dict = '{"fmap": true}'
    parsed_dict = read_user_dict(user_dict)

    assert isinstance(parsed_dict, dict)
    assert parsed_dict.get("fmap", None) is not None

    user_dict = '{"fmap": true : 2}'
    with pytest.raises(ValueError):
        _ = read_user_dict(user_dict)

    user_dict = '[{"fmap": true}]'
    with pytest.raises(ValueError):
        _ = read_user_dict(user_dict)

    sample_data = {"name": "test", "fmap": True, "input_model": "plexos"}

    yaml_file = tmp_path / "data.yaml"
    with yaml_file.open("w") as f:
        yaml.dump(sample_data, f)

    assert yaml_file.exists()
    parsed_dict = read_user_dict(str(yaml_file))
    assert isinstance(parsed_dict, dict)
    assert parsed_dict.get("fmap", None) is not None
    assert parsed_dict["input_model"] == "plexos"

    json_file = tmp_path / "data.json"
    with json_file.open("w") as f:
        json.dump(sample_data, f, indent=4)

    assert json_file.exists()
    parsed_dict = read_user_dict(str(json_file))
    assert isinstance(parsed_dict, dict)
    assert parsed_dict.get("fmap", None) is not None
    assert parsed_dict["input_model"] == "plexos"

    yaml_file = tmp_path / "missing_file.yaml"
    with pytest.raises(FileNotFoundError):
        _ = read_user_dict(str(yaml_file))


@pytest.mark.parametrize(
    "original, override, expected",
    [
        (
            {"bio_fuel_price": {"fname": "oldfile.csv", "old-key": True}},
            {"bio_fuel_price": {"fname": "repbioprice_2030.csv"}},
            {"bio_fuel_price": {"fname": "repbioprice_2030.csv", "old-key": True}},
        ),
        (
            {"bio_fuel_price": {"fname": "oldfile.csv"}},
            {"bio_fuel_price": {"fname": "repbioprice_2030.csv", "new_key": True}},
            {"bio_fuel_price": {"fname": "repbioprice_2030.csv", "new_key": True}},
        ),
        (
            {"plexos_device_map": {"old-key": True}},
            {"plexos_device_map": {"Lone Mountain": {"fuel": "GAS"}}},
            {"plexos_device_map": {"Lone Mountain": {"fuel": "GAS"}, "old-key": True}},
        ),
        (
            {"tech_to_fuel_pm": {"gas": {"fuel": "GAS", "type": "BA"}}},
            {"tech_to_fuel_pm": {"coal": {"fuel": None, "type": "BA"}}},
            {
                "tech_to_fuel_pm": {
                    "gas": {"fuel": "GAS", "type": "BA"},
                    "coal": {"fuel": None, "type": "BA"},
                }
            },
        ),
        (
            {"plugins": ["default_plugin"]},
            {"plugins": {"_replace": True, "list": ["break_gens", "pcm_defaults"]}},
            {"plugins": {"list": ["break_gens", "pcm_defaults"]}},
        ),
        (
            {"plexos_device_map": {"Old Plant": {"fuel": "COAL"}}},
            {"plexos_device_map": {"_replace": True, "Lone Mountain": {"fuel": "GAS"}}},
            {"plexos_device_map": {"Lone Mountain": {"fuel": "GAS"}}},
        ),
        (
            {"existing_key": {"sub_key": "value1"}},
            {"new_section": {"new_sub_key": "value2"}},
            {
                "existing_key": {"sub_key": "value1"},
                "new_section": {"new_sub_key": "value2"},
            },
        ),
        (
            {"key1": "value1"},
            None,
            {"key1": "value1"},
        ),
        (
            {"key1": {"value1": True}},
            {"key1": {"value1": False}},
            {"key1": {"value1": False}},
        ),
        (
            {"key1": "value1"},
            {},
            {"key1": "value1"},
        ),
    ],
    ids=[
        "override-existing",
        "override-existing-merge-new",
        "merge-new-keys-empty",
        "merge-new-keys-existing",
        "replace",
        "replace-nested",
        "replace-new-key",
        "no-override",
        "override-single-key",
        "full-replace-empty",
    ],
)
def test_update_dict(original, override, expected):
    result = override_dict(original, override)
    assert result == expected


def test_check_file_exist():
    csv_file_contents = "name,value\nTemp,25.5\nLoad,1200\nPressure,101.3"
    with NamedTemporaryFile(mode="w+", suffix=".csv", delete=False) as temp_file:
        temp_file.write(csv_file_contents)
        temp_file.seek(0)

        path = Path(temp_file.name)
        folder = path.parent
        fpath = check_file_exists(path.name, folder, folder=folder)
        assert path == fpath



from r2x.upgrader.helpers import get_function_arguments


def test_prepare_function_arguments():
    def dummy_function(a, b, c=None):
        pass

    data = {"a": 1, "b": 2, "c": 3, "extra": 4}
    result = get_function_arguments(data, dummy_function)
    expected = {"a": 1, "b": 2, "c": 3}
    assert result == expected, f"Expected {expected}, but got {result}"

    data = {"a": " 1 ", "b": " 2 "}
    result = get_function_arguments(data, dummy_function)
    expected = {"a": 1, "b": 2}
    assert result == expected, f"Expected {expected}, but got {result}"

    data = {"a": 1, "test": {"b": 1, "c": 2}}
    result = get_function_arguments(data, dummy_function)
    expected = {"a": 1, "b": 1, "c": 2}
    assert result == expected, f"Expected {expected}, but got {result}"

    data = {"a": 1, "test": '{"b": 1, "c": 2}'}
    result = get_function_arguments(data, dummy_function)
    expected = {"a": 1, "b": 1, "c": 2}
    assert result == expected, f"Expected {expected}, but got {result}"



"""Plugin to create time series for Imports.

This plugin create the time series representation for imports. Currently, it only process the canadian imports
on ReEDS.

This plugin is only applicable for ReEDs, but could work with similarly arrange data
"""

from datetime import datetime, timedelta

import polars as pl
from infrasys.time_series_models import SingleTimeSeries
from loguru import logger

from r2x.api import System
from r2x.config_models import ReEDSConfig
from r2x.config_scenario import Scenario
from r2x.models.generators import HydroDispatch
from r2x.parser.handler import BaseParser
from r2x.units import Energy


def update_system(
    config: Scenario,
    system: System,
    parser: BaseParser | None = None,
) -> System:
    """Apply an emission cap constraint for the system.

    This function adds to the sytem a :class:`~r2x.models.Constraint`object that is used to set the maximum
    emission per year.

    Parameters
    ----------
    config : Scenario
        The scenario configuration.
    parser : BaseParser
        The parser object used for parsing.
    system : System
        The system object to be updated.
    """
    assert config.input_config
    if not isinstance(config.input_config, ReEDSConfig):
        msg = "Plugin `imports.py` is not compatible with an input model that is not ReEDS."
        raise NotImplementedError(msg)
    assert config.input_config.weather_year
    weather_year = config.input_config.weather_year

    if parser is None:
        msg = "Missing parser information for imports. Skipping plugin."
        logger.debug(msg)
        return system

    if parser is not None:
        # NOTE: We might change this condition once we change the imports definition on ReEDS.
        assert all(key in parser.data for key in ["canada_imports", "canada_szn_frac"]), (
            "Missing required files for import plugin."
        )
        assert "hour_map" in parser.data, "Missing hour map from ReEDS run."

    logger.info("Adding imports time series...")
    hour_map = parser.data["hour_map"]
    szn_frac = parser.data["canada_szn_frac"]
    total_imports = parser.data["canada_imports"]

    hourly_time_series = hour_map.join(szn_frac, on="season", how="left")
    assert not hourly_time_series.is_empty()
    hourly_time_series = hourly_time_series.with_columns(
        pl.col("time_index").str.to_datetime(),
    )
    daily_time_series = hourly_time_series.group_by(pl.col("time_index").dt.date()).median()

    # NOTE: Since the seasons can be repeated, the szn frac can be greater than one. To avoid this, we
    # normalize it again to redistributed the fraction throught the 365 or 366 days.
    daily_time_series_normalized = daily_time_series.with_columns(pl.col("value") / pl.col("value").sum())

    # NOTE: This will need change if we modify the model for the imports. CUrrently all is assumed to be
    # modeled as HydroEnergyReservoir. Currently we only apply it to can-imports.
    initial_time = datetime(year=weather_year, month=1, day=1)
    for generator in system.get_components(HydroDispatch, filter_func=lambda x: "can-imports" in x.name):
        daily_budget = (
            total_imports.filter(pl.col("r") == generator.bus.name)["value"].item()
            * daily_time_series_normalized["value"].to_numpy()
        )
        ts = SingleTimeSeries.from_array(
            data=Energy(daily_budget[:-1] / 1e3, "GWh"),
            variable_name="hydro_budget",
            initial_time=initial_time,
            resolution=timedelta(days=1),
        )
        system.add_time_series(ts, generator)
    return system



"""Configuration file for the Sphinx documentation builder."""

project = "R2X"
copyright = "2024, Alliance for Sustainable Energy, LLC"
author = "R2X developers"

# Add more extensions here
extensions = [
    "myst_parser",
    "sphinx_copybutton",
    "sphinx.ext.autodoc",
    "sphinx_tabs.tabs",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinxcontrib.mermaid",
    "sphinxcontrib.autodoc_pydantic",
]
# Make sure the target is unique
autosectionlabel_prefix_document = True

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# autosummary_generate = True  # Turn on sphinx.ext.autosummary


# Adding other sphinx documentation as hyperlinks
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pandas": ("https://pandas.pydata.org/docs", None),
    "infrasys": ("https://nrel.github.io/infrasys/", None),
}

html_theme = "sphinx_book_theme"
html_static_path = ["_static"]
source_suffix = [".md"]

html_theme_options = {
    "repository_url": "https://github.com/NREL/R2X",
    "path_to_docs": "docs/source/",
    "show_toc_level": 3,
    "use_source_button": True,
    "use_edit_page_button": True,
}

myst_enable_extensions = [
    "amsmath",
    "attrs_inline",
    "attrs_block",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "colon_fence",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]

# Pydantic stuff
autodoc_pydantic_model_show_config_summary = False

# -- Options for autodoc ----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#configuration

# Automatically extract typehints when specified and place them in
# descriptions of the relevant function/method.
autodoc_typehints = "description"

# Don't show class signature with the class' name.
# autodoc_class_signature = "separated"
suppress_warnings = ["myst.header"]

# Copy button
copybutton_exclude = ".linenos, .gp, .go"



# ruff: noqa
from .attributes import Emission
from .branch import (
    ACBranch,
    AreaInterchange,
    Branch,
    DCBranch,
    Line,
    MonitoredLine,
    PhaseShiftingTransformer,
    TapTransformer,
    TModelHVDCLine,
    Transformer2W,
    TwoTerminalHVDCLine,
)
from .core import MinMax, ReserveMap, TransmissionInterfaceMap
from .costs import HydroGenerationCost, RenewableGenerationCost, StorageCost, ThermalGenerationCost
from .generators import (
    EnergyReservoirStorage,
    Generator,
    EnergyReservoirStorage,
    HybridSystem,
    HydroDispatch,
    HydroEnergyReservoir,
    HydroGen,
    HydroPumpedStorage,
    RenewableDispatch,
    RenewableGen,
    RenewableNonDispatch,
    Storage,
    ThermalGen,
    ThermalStandard,
)
from .load import FixedAdmittance, InterruptiblePowerLoad, PowerLoad, StandardLoad
from .services import Reserve, TransmissionInterface, VariableReserve
from .topology import ACBus, Arc, Area, Bus, DCBus, LoadZone
from .utils import GeoLocation, GeographicInfo



import pytest

from r2x.config_models import BaseModelConfig, Models, ReEDSConfig
from r2x.config_utils import (
    get_input_defaults,
    get_input_model_fmap,
    get_model_config_class,
    get_output_defaults,
)


def test_get_input_defaults():
    model = Models.REEDS

    defaults = get_input_defaults(model)
    assert isinstance(defaults, dict)

    with pytest.raises(ValueError):
        _ = get_input_defaults(Models.PRAS)


def test_get_output_defaults():
    model = Models.PLEXOS
    defaults = get_output_defaults(model)
    assert isinstance(defaults, dict)

    with pytest.raises(ValueError):
        _ = get_output_defaults(Models.REEDS)


def test_get_input_model_fmap():
    model = Models.REEDS
    defaults = get_input_model_fmap(model)
    assert isinstance(defaults, dict)

    with pytest.raises(ValueError):
        _ = get_input_model_fmap(Models.PRAS)


def test_get_model_config_class():
    model = Models.REEDS

    config = get_model_config_class(model)
    assert isinstance(config, BaseModelConfig)
    assert isinstance(config, ReEDSConfig)



import pytest
import polars as pl
from pathlib import Path
from polars.testing import assert_frame_equal
from tempfile import NamedTemporaryFile
from r2x.parser.handler import csv_handler
from r2x.parser.plexos_utils import find_xml


@pytest.fixture
def sample_csv_basic():
    data = "ID,Name,Age\n1,Alice,30\n2,Bob,24"
    return data


@pytest.fixture
def temp_csv_file(sample_csv_basic):
    with NamedTemporaryFile(mode="w", delete=False, suffix=".csv") as temp_file:
        temp_file.write(sample_csv_basic)
        temp_file.flush()
        return Path(temp_file.name)


def test_csv_handler_basic(temp_csv_file):
    """Test if csv_handler correctly reads a CSV and converts column names to lowercase."""
    fpath = temp_csv_file

    df_test = csv_handler(fpath)
    assert df_test is not None

    expected_df = pl.DataFrame({"id": [1, 2], "name": ["alice", "bob"], "age": [30, 24]})

    assert_frame_equal(df_test, expected_df)

    with pytest.raises(FileNotFoundError):
        _ = csv_handler(Path("non_existent_file.csv"))


@pytest.fixture
def basic_xml():
    data = """
            <t_category>
              <category_id>1</category_id>
              <class_id>1</class_id>
              <rank>0</rank>
              <name>-</name>
            </t_category>
           """
    return data


def test_find_xml(tmp_path: Path, basic_xml: str):
    # Case 1: no xml files in tmp_path
    with pytest.raises(FileNotFoundError):
        find_xml(tmp_path)

    # Case 2: 1 xml file in tmp_path
    xml_file_1 = tmp_path / "file1.xml"
    xml_file_1.write_text(basic_xml)

    assert xml_file_1 == find_xml(tmp_path)

    # Case 3: 2 xml files in tmp_path
    xml_file_2 = tmp_path / "file2.xml"
    xml_file_2.write_text(basic_xml)

    with pytest.raises(FileNotFoundError):
        find_xml(tmp_path)



import pytest
from r2x.api import System
from r2x.config_scenario import Scenario
from r2x.models.branch import MonitoredLine
from r2x.models.topology import ACBus, LoadZone
from r2x.plugins.hurdle_rate import update_system
from r2x.runner import run_parser


def test_hurdle_rate():
    system = System(name="TestSystem")
    region_1 = LoadZone(name="region_1")
    region_2 = LoadZone(name="region_2")
    system.add_components(region_1, region_2)

    bus_1 = ACBus(number=1, name="Bus1", load_zone=region_1)
    bus_2 = ACBus(number=2, name="Bus2", load_zone=region_1)
    bus_3 = ACBus(number=3, name="Bus3", load_zone=region_2)
    system.add_components(bus_1, bus_2, bus_3)

    # Line inside regions
    line_1_2 = MonitoredLine(
        name="1-2",
        from_bus=bus_1,
        to_bus=bus_2,
        ext={"Wheeling Charge": 0.001, "Wheeling Charge Back": 0.001},
    )
    # Line inside regions
    line_2_3 = MonitoredLine(name="2-3", from_bus=bus_2, to_bus=bus_3)
    # Line between regions
    line_1_3 = MonitoredLine(
        name="1-3",
        from_bus=bus_1,
        to_bus=bus_3,
        ext={"Wheeling Charge": 0.001, "Wheeling Charge Back": 0.001},
    )
    system.add_components(line_1_2, line_2_3, line_1_3)

    config = Scenario.from_kwargs(
        name="5bus",
        input_model="reeds-US",
        output_model="plexos",
        solve_year=2035,
        weather_year=2012,
    )
    hurdle_rate_value = 0.006
    new_system = update_system(config=config, system=system, hurdle_rate=hurdle_rate_value)
    assert isinstance(new_system, System)

    # Check line between regions
    updated_line_1_3 = system.get_component(MonitoredLine, name="1-3")
    assert isinstance(updated_line_1_3, MonitoredLine)
    assert updated_line_1_3.ext.get("Wheeling Charge")
    assert updated_line_1_3.ext["Wheeling Charge"] == hurdle_rate_value

    # Check line inside regions did not changed
    updated_line_1_2 = system.get_component(MonitoredLine, name="1-2")
    assert isinstance(updated_line_1_3, MonitoredLine)
    assert updated_line_1_2.ext.get("Wheeling Charge")
    assert updated_line_1_2.ext["Wheeling Charge"] == 0.001

    # Test invalid models
    config.output_model = "sienna"
    with pytest.raises(NotImplementedError):
        _ = update_system(config=config, system=system, hurdle_rate=hurdle_rate_value)


def test_hurdle_rate_with_parser(reeds_data_folder, tmp_folder):
    config = Scenario.from_kwargs(
        name="5bus",
        run_folder=reeds_data_folder,
        output_folder=tmp_folder,
        input_model="reeds-US",
        output_model="plexos",
        solve_year=2035,
        weather_year=2012,
    )

    system, parser = run_parser(config)

    hurdle_rate_value = 0.006
    new_system = update_system(config=config, parser=parser, system=system, hurdle_rate=hurdle_rate_value)
    assert isinstance(new_system, System)



"""Testing for the configuration and Scenario class."""

import pytest

from r2x.config_models import PlexosConfig, ReEDSConfig, SiennaConfig
from r2x.config_scenario import Configuration, Scenario, get_scenario_configuration
from r2x.utils import read_fmap


@pytest.fixture
def scenario(data_folder, tmp_folder):
    return Scenario(
        name="Test Scenario",
        run_folder=data_folder,
        output_folder=tmp_folder,
        input_model="plexos",
        output_model="sienna",
    )


def test_scenario_instance(scenario, data_folder, tmp_folder):
    assert isinstance(scenario, Scenario)
    assert scenario.name == "Test Scenario"
    assert scenario.input_model == "plexos"
    assert isinstance(scenario.input_config, PlexosConfig)
    assert scenario.output_model == "sienna"
    assert isinstance(scenario.output_config, SiennaConfig)
    assert scenario.run_folder == data_folder
    assert scenario.output_folder == tmp_folder


def test_scenario_mkdirs(scenario, tmp_folder):
    assert scenario.output_folder.exists()
    assert scenario.output_folder == tmp_folder


@pytest.mark.parametrize(
    "input_model, output_model, expected_fmap",
    [
        ("reeds-US", "plexos", read_fmap("r2x/defaults/reeds_us_mapping.json")),
        ("reeds-US", "sienna", read_fmap("r2x/defaults/reeds_us_mapping.json")),
        ("plexos", "sienna", read_fmap("r2x/defaults/plexos_mapping.json")),
        ("sienna", "plexos", read_fmap("r2x/defaults/sienna_mapping.json")),
    ],
    ids=["R2P", "R2S", "P2S", "SP"],
)
def test_scenario_fmap(input_model, output_model, expected_fmap):
    scenario = Scenario.from_kwargs(
        name=f"test-{input_model}", input_model=input_model, output_model=output_model
    )
    assert scenario.input_config
    assert scenario.input_config.fmap == expected_fmap


def test_scenario_from_kwargs(tmp_folder):
    kwargs = {
        "name": "Test",
        "weather_year": 2015,
        "solve_year": 2055,
        "output_folder": tmp_folder,
        "input_model": "reeds-US",
        "output_model": "sienna",
        "feature_flags": {"cool-feature": True},
    }
    scenario = Scenario.from_kwargs(**kwargs)
    assert isinstance(scenario, Scenario)
    assert scenario.input_model == "reeds-US"
    assert scenario.output_model == "sienna"
    assert scenario.feature_flags
    assert scenario.feature_flags.get("cool-feature", None)
    assert scenario.input_config
    assert hasattr(scenario.input_config, "weather_year")
    assert getattr(scenario.input_config, "weather_year", None) == 2015
    assert hasattr(scenario.input_config, "solve_year")
    assert getattr(scenario.input_config, "solve_year", None) == 2055


#
def test_config_from_cli():
    cli_input = {
        "name": "test",
        "weather_year": 2015,
        "solve_year": 2055,
        "input_model": "plexos",
        "output_model": "sienna",
        "feature_flags": {"cool-feature": True},
    }
    config = Configuration.from_cli(cli_args=cli_input)

    assert isinstance(config, Configuration)
    assert len(config) == 1
    assert config["test"].input_config
    assert isinstance(config["test"].input_config, PlexosConfig)

    user_dict = {"fmap": {"xml_file": {"fname": "path.xml", "another_field": True}}}
    config = Configuration.from_cli(cli_args=cli_input, user_dict=user_dict)
    assert isinstance(config, Configuration)
    assert len(config) == 1
    scenario = config["test"]
    assert scenario.input_config
    assert scenario.input_config.fmap["xml_file"]["fname"] == user_dict["fmap"]["xml_file"]["fname"]
    assert scenario.input_config.fmap["xml_file"]["another_field"]


def test_config_from_scenarios():
    user_dict = {
        "input_model": "reeds-US",
        "output_model": "sienna",
        "scenarios": [
            {
                "name": "test2030",
                "weather_year": 2015,
                "solve_year": 2030,
            },
            {
                "name": "test2050",
                "weather_year": 2015,
                "solve_year": 2055,
            },
        ],
    }

    config = Configuration.from_scenarios({}, user_dict)
    assert config is not None
    assert len(config) == 2
    assert config["test2030"]
    assert config["test2030"].input_config
    assert config["test2030"].input_config
    assert hasattr(config["test2030"].input_config, "weather_year")
    assert getattr(config["test2030"].input_config, "weather_year") == 2015


@pytest.mark.parametrize(
    "cli_input,user_dict",
    [
        (
            {
                "name": "Test",
                "weather_year": 2015,
                "solve_year": [2055],
                "input_model": "plexos",
                "output_model": "sienna",
            },
            {},
        ),
        (
            {},
            {
                "input_model": "reeds-US",
                "output_model": "sienna",
                "scenarios": [
                    {
                        "name": "test2030",
                        "weather_year": 2015,
                        "solve_year": 2030,
                    },
                    {
                        "name": "test2050",
                        "weather_year": 2015,
                        "solve_year": 2055,
                    },
                ],
            },
        ),
        (
            {
                "name": "Test",
                "weather_year": 2015,
                "input_model": "plexos",
                "output_model": "sienna",
            },
            {
                "scenarios": [
                    {
                        "name": "test2030",
                        "solve_year": 2030,
                    },
                    {
                        "name": "test2050",
                        "solve_year": 2055,
                    },
                ],
            },
        ),
    ],
    ids=["no-user-dict", "no-cli", "both"],
)
def test_get_config(cli_input, user_dict):
    config = get_scenario_configuration(cli_input, user_dict)
    assert config is not None


def test_get_config_cli_override():
    cli_args = {"name": "TestConfig", "input_model": "reeds-US", "output_model": "plexos"}
    user_dict = {"input_model": "plexos"}
    config = get_scenario_configuration(cli_args, user_dict)
    assert config is not None
    assert isinstance(config, Configuration)
    assert config["TestConfig"].input_model == "reeds-US"


def test_update_configuration():
    cli_args = {}

    new_tech_key = "new_tech"
    new_tech = {new_tech_key: {"fuel": None, "type": "HY"}}
    user_dict = {
        "name": "test",
        "tech_to_fuel_pm": new_tech,
        "input_model": "reeds-US",
        "output_model": "plexos",
    }
    config = get_scenario_configuration(cli_args, user_dict)
    assert new_tech_key in config["test"].input_config.defaults["tech_to_fuel_pm"]
    assert config["test"].input_config.defaults["tech_to_fuel_pm"][new_tech_key] == new_tech[new_tech_key]


def test_update_defaults():
    cli_args = {}

    default_key = "commit_technologies"
    default_values = [0, 1, 2]
    user_dict = {
        "name": "test",
        default_key: default_values,
        "input_model": "reeds-US",
        "output_model": "plexos",
    }
    config = get_scenario_configuration(cli_args, user_dict)
    assert config["test"].input_config.defaults[default_key] == default_values

    cli_args = {}

    default_key = "commit_technologies"
    default_values = [0, 1, 2]
    user_dict = {
        "name": "test",
        default_key: default_values,
        "input_model": "reeds-US",
        "scenarios": [{"name": "Test1", "solve_year": 2030}, {"name": "Test2", "solve_year": 2040}],
        "output_model": "plexos",
    }
    config = get_scenario_configuration(cli_args, user_dict)
    for scenario_name, scenario_config in config:
        assert isinstance(scenario_config, Scenario)
        assert isinstance(scenario_config.input_config, ReEDSConfig)
        assert scenario_config.input_config.defaults[default_key] == default_values


def test_correct_update_fmap():
    cli_args = {"upgrade": True}
    user_dict = {
        "name": "test",
        "input_model": "reeds-US",
        "output_model": "plexos",
        "fmap": {"online_capacity": {"fname": "test", "folder": True}},
    }
    config = get_scenario_configuration(cli_args, user_dict)
    scenario_config = config["test"]
    assert isinstance(scenario_config, Scenario)
    assert isinstance(scenario_config.input_config, ReEDSConfig)
    assert len(scenario_config.input_config.fmap) > 1
    assert (
        scenario_config.input_config.fmap["online_capacity"]["fname"]
        == user_dict["fmap"]["online_capacity"]["fname"]
    )
    assert (
        scenario_config.input_config.fmap["online_capacity"]["folder"]
        == user_dict["fmap"]["online_capacity"]["folder"]
    )

    cli_args = {}
    user_dict = {
        "name": "test",
        "input_model": "reeds-US",
        "output_model": "plexos",
        "scenarios": [{"name": "Test1", "solve_year": 2030}, {"name": "Test2", "solve_year": 2040}],
        "fmap": {"online_capacity": {"fname": "test", "folder": True}},
    }
    config = get_scenario_configuration(cli_args, user_dict)
    for scenario_name, scenario_config in config:
        assert isinstance(scenario_config, Scenario)
        assert isinstance(scenario_config.input_config, ReEDSConfig)
        assert (
            scenario_config.input_config.fmap["online_capacity"]["fname"]
            == user_dict["fmap"]["online_capacity"]["fname"]
        )
        assert (
            scenario_config.input_config.fmap["online_capacity"]["folder"]
            == user_dict["fmap"]["online_capacity"]["folder"]
        )


def test_configuration_printing(scenario):
    scenario.info()



from pathlib import Path
from tempfile import NamedTemporaryFile

import pytest

from r2x.parser.handler import file_handler


def test_file_handler():
    file_content = "name,value\nTemp,25.5\nLoad,1200\nPressure,101.3"
    with NamedTemporaryFile(mode="w+", suffix=".asd", delete=False) as temp_file:
        temp_file.write(file_content)
        temp_file.seek(0)

        with pytest.raises(NotImplementedError):
            _ = file_handler(Path(temp_file.name))
